/**
 * Introduction
 *
 * @title ROV - Easy applications from C
 *
 * @copyright Jason Tribbeck / 7th ARM 1994-2001
 *
 * @authors Jason Tribbeck
 *
 * @file appquit.c
 */

#include <stdio.h>
#include <stdarg.h>
#include "kernel.h"
#include <string.h>

#ifndef __GNUC__
#pragma -s1
#endif

#include "main.h"
#include "canvas.h"
#include "icons.h"
#include "file.h"
#include "msg.h"
#include "draw.h"

typedef void(* appquit_handler)(void);

static canvas appquit_canvas;
static icon _i[3];
static char _txt[256];
static int files = 0, quit_fromprogram = 0, sender;
static appquit_handler _quit = NULL;

#define APPQUIT_WIDTH  (512+64)
#define APPQUIT_HEIGHT (288)

void appquit_quit(void)
{
  quit_fromprogram = 1;

  if((poll_lastevent == 17) || (poll_lastevent == 18)) {
    if(poll_block.action == 8) {
      quit_fromprogram = 0;
    }
  }

  if(files) {
    if(files == 1) {
      sprintf(_txt, msg_lookup("AppQuit1:You have 1 file modified. Click on Cancel to abort quit, or Discard to discard the file."));
    } else {
      sprintf(_txt, msg_lookup("AppQuit2:You have %d files modified. Click on Cancel to abort quit, or Discard to discard the files."), files);
    }
    canvas_opencentre(appquit_canvas);
    if(!quit_fromprogram) {
      sender = poll_block.sender;
      if((poll_block.size > 20) && ((poll_block.data.words[0] & 1) == 1)) quit_fromprogram = 1;
      main_acknowledge(& poll_block, 19, 8);
    }
    canvas_claimcaret(appquit_canvas);
  } else {
    main_exit();
    if(_quit) _quit();
    exit(0);
  }
}

static void _appquit_close(void)
{
  canvas_close(appquit_canvas);
}

static void _appquit_closedown(void)
{
  if(!quit_fromprogram) {
    canvas_keystr c;
    canvas_getcaretstate((canvas_caretstr *) & c);
    c.chcode = 0x1fc;
    main_send((message_block *) & c, 8, sender, 0);
  }
  main_exit();
  if(_quit) _quit();
  exit(0);
}

static void _appquit_key(canvas c, canvas_keystr * k)
{
  if((k -> chcode < 32) || (k -> chcode > 255)) {
    canvas_close(appquit_canvas);
    main_processkey(k -> chcode);
  }
}

void appquit_setfiles(int n)
{
  files = n;
}

void appquit_seteditedfiles(int n)
{
  files = n;
}

void appquit_addfiles(int n)
{
  files += n;
}

void appquit_addeditedfiles(int n)
{
  files = n;
}

void appquit_init(appquit_handler _q)
{
  icon_data d;

  appquit_canvas = canvas_create(APPQUIT_WIDTH, APPQUIT_HEIGHT, TASKNAME);
  if(!appquit_canvas) {
    simple_warning("Unable to create appquit canvas", 1);
    exit(0);
  }
  canvas_setflags(appquit_canvas, 0x7f000000U, CANVAS_TITLEBAR);
  canvas_setcaretcolour(appquit_canvas, 2);
  canvas_make(appquit_canvas);
  d.indirecttext.text = _txt;
  d.indirecttext.valid = "l40";
  d.indirecttext.length = 256;
  _i[0] = icon_create(appquit_canvas, 16, 88, APPQUIT_WIDTH - 16, APPQUIT_HEIGHT - 16, 0x17000119, & d);
  _i[1] = icon_createbutton(appquit_canvas, APPQUIT_WIDTH * 1 / 4 - 80, 16, msg_lookuprom("Discard:Discard"), 160);
  _i[2] = icon_createdefaultbutton(appquit_canvas, APPQUIT_WIDTH * 3 / 4 - 80, 16, msg_lookuprom("Cancel:Cancel"), 160);
  main_addhandler(HANDLER_PREQUIT, appquit_quit);
  if(_q) _quit = _q;
  icon_addhandler(_i[2], ICON_CLICKHANDLER, _appquit_close);
  icon_addhandler(_i[1], ICON_CLICKHANDLER, _appquit_closedown);
  canvas_addhandler(appquit_canvas, CANVAS_KEYHANDLER, _appquit_key);

  return;
}
