/**
 * Introduction
 *
 * @title ROV - Easy applications from C
 *
 * @copyright Jason Tribbeck / 7th ARM 1994-2001
 *
 * @authors Jason Tribbeck
 *
 * @file newmalloc.c
 */

/* Uncomment out if you want debugging enabled
#define DEBUG1
*/

#include <stdio.h>

#include "kernel.h"

#include "main.h"
#include "genheap.h"
#include "extras.h"

typedef int (* rmalloc_extendp)(int);
typedef void * (* rmalloc_initp)(void);

static rmalloc_extendp rmext = NULL;
static rmalloc_initp rminit = NULL;

static void * rmalloc_memory = NULL;
static int rmalloc_size;

#ifdef DEBUG1
void rmalloc_crash(void)
{
}
#endif

char * rmalloc_alloc(int size)
{
  void * a;
  int ns;

  size = (size + 3) & 0xfffffffc;

  if(!rmalloc_memory) {
    if(!rminit) {
      swerr(TRUE, "rmalloc_new not initialised properly");
    }
    rmalloc_memory = rminit();
    rmalloc_size = rmext(256);
    genheap_init(rmalloc_memory, rmalloc_size);
  }

#ifdef DEBUG1
  swerr(FALSE, "Allocating &%-.8x bytes", size);
  if(check_key(0)) {
    FILE * f;
    void * ptr;
    int n;

    f = fopen("RAM:$.AllocTree", "w");
    ptr = 0;
    n = 0;
    fprintf(f, "Start memory:&%-.8x\n", (int)rmalloc_memory);
    fprintf(f, "Memory size :&%-.8x\n", rmalloc_size);
    fprintf(f, "Top         :&%-.8x\n", ((int)rmalloc_memory) + rmalloc_size);
    do {
      if(ptr) fprintf(f, "%-.8x [%c] [%-.8x]\n", (int)ptr, genheap_type(ptr), genheap_size(ptr));
      ptr = genheap_walk(rmalloc_memory, rmalloc_size, ptr);
      n ++ ;
    } while((ptr) && (n < 1024));
    if(n == 1024) {
      fprintf(f, ">> ABORTED [cyclic] <<\n");
    }
    fclose(f);
    swerr(FALSE, "Map made!");
  }
#endif
  a = genheap_malloc(rmalloc_memory, rmalloc_size, size);
#ifdef DEBUG1
  if(a) swerr(FALSE, "Allocated %x at &%-.8x", size, (int)a);
#endif
  if(a) return (char *)a;

  ns = rmext(rmalloc_size + size + 16);
  if(ns > rmalloc_size) {
    genheap_resize(rmalloc_memory, rmalloc_size, ns);
    rmalloc_size = ns;
  }
  a = genheap_malloc(rmalloc_memory, rmalloc_size, size);
#ifdef DEBUG1
  swerr(FALSE, "Allocated %x at &%-.8x [reallocated to %-.8x]", size, (int)a, rmalloc_size);
#endif

  return (char *)a;
}

void rmalloc_free(char * a)
{
  if(!rmalloc_memory) {
    return;
  }
#ifdef DEBUG1
  swerr(FALSE, "Freed &%-.8x", (int)a);
#endif
  genheap_free(rmalloc_memory, rmalloc_size, a);
  return;
}

static char * rcalloc_alloc(int a, int b)
{
  return rmalloc_alloc(a * b);
}

void rmalloc_newinit(void)
{
  main_addhandler(HANDLER_RMALLOC, rmalloc_alloc);
  main_addhandler(HANDLER_RCALLOC, rcalloc_alloc);
  main_addhandler(HANDLER_RFREE, rmalloc_free);
}

void rmalloc_claimhandlers(rmalloc_extendp e, rmalloc_initp i)
{
  rmext = e;
  rminit = i;
  /* first_object = i(); */
}
