/**
 * Introduction
 *
 * @title ROV - Easy applications from C
 *
 * @copyright Jason Tribbeck / 7th ARM 2002
 *
 * @authors Jason Tribbeck
 *
 * @file omemcpy.c
 */

#include <stdlib.h>
#include <string.h>

extern void swerr(int, char* b, ...);

void omemcpy(char *dst, char *src, int length)
{
  if((length <= 0) || (src == dst)) {
    return;
  }

  if(dst > src) {
    if(dst >= src + length) {
      memcpy(dst, src, length);
      return;
    } else {
#ifdef USE_RECURSION
      int overlap = (int)src + length - (int)dst;
      int underlap = length - overlap;
      omemcpy(dst + underlap, src + underlap, overlap);
      omemcpy(dst, src, underlap);
#else
      //swerr(0, "dst is higher than src [%p, %p, %-.8x]", dst, src, length);
      // dst is higher than src, so we need to copy top to bottom
      if((((int)dst | (int)src | length) & 3) == 0) {
        // it's all word aligned
        unsigned int* dsti = (unsigned int*)dst;
        unsigned int* srci = (unsigned int*)src;

        length = length >> 2;
        for(length--; length >= 0; length--) {
          dsti[length] = srci[length];
        }
      } else {
        for(length--; length >= 0; length--) {
          dst[length] = src[length];
        }
      }
#endif
    }
  } else {
    if(src >= dst + length) {
      memcpy(dst, src, length);
      return;
    } else {
#ifdef USE_RECURSION
      int overlap = (int)dst + length - (int)src;
      int underlap = length - overlap;
      omemcpy(dst, src, overlap);
      omemcpy(dst + overlap, src + overlap, underlap);
#else
      //swerr(0, "src is higher than dst [%p, %p, %-.8x]", dst, src, length);
      // src is higher than dst, so we need to copy bottom to top
      if((((int)dst | (int)src | length) & 3) == 0) {
        // it's all word aligned
        unsigned int* dsti = (unsigned int*)dst;
        unsigned int* srci = (unsigned int*)src;

        length = length >> 2;
        for(length--; length >= 0; length--) {
          *dsti = *srci;
          dsti++;
          srci++;
        }
      } else {
        for(length--; length >= 0; length--) {
          *dst = *src;
          dst++;
          src++;
        }
      }
#endif
    }
  }
}
